package loosy

import (
	"fmt"
	"strconv"
	"strings"
	"unsafe"

	jsoniter "github.com/json-iterator/go"
	"github.com/modern-go/reflect2"
)

func init() {
	// register the JSON{} json encoder / decoders.
	jsonType := reflect2.TypeOf(JSON{}).String()
	jsoniter.RegisterTypeEncoder(jsonType, jsonencoder{})
	jsoniter.RegisterTypeDecoder(jsonType, jsondecoder{})

	// register the Any{} json encoder / decoders.
	anyType := reflect2.TypeOf(Any{}).String()
	jsoniter.RegisterTypeEncoder(anyType, anyencoder{})
	jsoniter.RegisterTypeDecoder(anyType, anydecoder{})

	// register the []*JSON{} json encoder / decoders.
	jsonArrType := reflect2.TypeOf([]*JSON{}).String()
	jsoniter.RegisterTypeEncoder(jsonArrType, jsonarrayencoder{})
	jsoniter.RegisterTypeDecoder(jsonArrType, jsonarraydecoder{})

	// register the []*Any{} json encoder / decoders.
	anyArrType := reflect2.TypeOf([]*Any{}).String()
	jsoniter.RegisterTypeEncoder(anyArrType, anyarrayencoder{})
	jsoniter.RegisterTypeDecoder(anyArrType, anyarraydecoder{})

	// register the map[string]*JSON{} json encoder / decoders.
	jsonObjType := reflect2.TypeOf(map[string]*JSON{}).String()
	jsoniter.RegisterTypeEncoder(jsonObjType, jsonobjectencoder{})
	jsoniter.RegisterTypeDecoder(jsonObjType, jsonobjectdecoder{})

	// register the map[string]*Any{} json encoder / decoders.
	anyObjType := reflect2.TypeOf(map[string]*Any{}).String()
	jsoniter.RegisterTypeEncoder(anyObjType, anyobjectencoder{})
	jsoniter.RegisterTypeDecoder(anyObjType, anyobjectdecoder{})
}

// jsonencoder implements jsoniter.ValEncoder{} for JSON{}.
type jsonencoder struct{}

func (_ jsonencoder) IsEmpty(ptr unsafe.Pointer) bool {
	j := (*JSON)(ptr)
	return j.IsEmpty()
}

func (_ jsonencoder) Encode(ptr unsafe.Pointer, stream *jsoniter.Stream) {
	encodeJSON(stream, (*JSON)(ptr))
}

// jsondecoder implements jsoniter.ValDecoder{} for JSON{}.
type jsondecoder struct{}

func (_ jsondecoder) Decode(ptr unsafe.Pointer, iter *jsoniter.Iterator) {
	decodeJSON(iter, (*JSON)(ptr))
}

// encodeJSON encodes data in JSON{} into Stream{}.
func encodeJSON(stream *jsoniter.Stream, j *JSON) {
	stream.Write(j.B)
}

// decodeJSON decodes next available value in Iterator{} into JSON{}.
func decodeJSON(iter *jsoniter.Iterator, j *JSON) {
	j.B = iter.SkipAndReturnBytes()
}

// jsonarrayencoder implements jsoniter.ValEncoder{} for []*JSON{}.
type jsonarrayencoder struct{}

func (_ jsonarrayencoder) IsEmpty(ptr unsafe.Pointer) bool {
	return *(*[]*JSON)(ptr) == nil
}

func (_ jsonarrayencoder) Encode(ptr unsafe.Pointer, stream *jsoniter.Stream) {
	encodeJSONArray(stream, *(*[]*JSON)(ptr))
}

// jsonarraydecoder implements jsoniter.ValDecoder{} for []*JSON{}.
type jsonarraydecoder struct{}

func (_ jsonarraydecoder) Decode(ptr unsafe.Pointer, iter *jsoniter.Iterator) {
	decodeJSONArray(iter, (*[]*JSON)(ptr))
}

// encodeJSONArray encodes data in []*JSON{} into Stream{}.
func encodeJSONArray(stream *jsoniter.Stream, arr []*JSON) {
	if arr == nil {
		// Write nil value.
		stream.WriteNil()
		return
	}

	stream.WriteArrayStart()
	for i := 0; i < len(arr)-1; i++ {
		encodeJSON(stream, arr[i])
		stream.WriteMore()
	}
	encodeJSON(stream, arr[len(arr)-1])
	stream.WriteArrayEnd()
}

// decodeJSONArray decodes data in Iterator{} into []*JSON{}.
func decodeJSONArray(iter *jsoniter.Iterator, arr *[]*JSON) {
	if arr == nil {
		// the ptr itself should never be nil.
		panic("BUG: decoder nil unsafe.Pointer")
	}

	switch iter.WhatIsNext() {
	// Allocate new arry slice.
	case jsoniter.ArrayValue:
		*arr = make([]*JSON, 32)

	// Don't set on nil value.
	case jsoniter.NilValue:
		_ = iter.ReadNil()
		return

	// Unusable type.
	default:
		iter.ReportError("decodeJSONArray", "invalid type")
		iter.Skip() // skip this array.
		return
	}

	for iter.ReadArray() {
		elem := new(JSON)
		decodeJSON(iter, elem)
		(*arr) = append((*arr), elem)
	}
}

// jsonobjectencoder implements jsoniter.ValEncoder{} for map[string]*JSON{}.
type jsonobjectencoder struct{}

func (_ jsonobjectencoder) IsEmpty(ptr unsafe.Pointer) bool {
	return *(*map[string]*JSON)(ptr) == nil
}

func (_ jsonobjectencoder) Encode(ptr unsafe.Pointer, stream *jsoniter.Stream) {
	encodeJSONObject(stream, *(*map[string]*JSON)(ptr))
}

// jsonobjectdecoder implements jsoniter.ValDecoder{} for map[string]*JSON{}.
type jsonobjectdecoder struct{}

func (_ jsonobjectdecoder) Decode(ptr unsafe.Pointer, iter *jsoniter.Iterator) {
	decodeJSONObject(iter, (*map[string]*JSON)(ptr))
}

// encodeJSONObject encodes data in map[string]*JSON{} into Stream{}.
func encodeJSONObject(stream *jsoniter.Stream, obj map[string]*JSON) {
	if obj == nil {
		// Write nil value.
		stream.WriteNil()
		return
	}

	var more bool

	stream.WriteObjectStart()
	for key, val := range obj {
		if more {
			stream.WriteMore()
		}
		stream.WriteObjectField(key)
		encodeJSON(stream, val)
		more = true
	}
	stream.WriteObjectEnd()
}

// decodeJSONObject decodes data in Iterator{} into map[string]*JSON{}.
func decodeJSONObject(iter *jsoniter.Iterator, obj *map[string]*JSON) {
	if obj == nil {
		// the ptr itself should never be nil.
		panic("BUG: decoder nil unsafe.Pointer")
	}

	switch iter.WhatIsNext() {
	// Allocate new object map.
	case jsoniter.ObjectValue:
		*obj = make(map[string]*JSON)

	// Don't set on nil value.
	case jsoniter.NilValue:
		_ = iter.ReadNil()
		return

	// Unusable type.
	default:
		iter.ReportError("decodeJSONObject", "invalid type")
		iter.Skip() // skip this object.
		return
	}

	for {
		key := iter.ReadObject()
		if key == "" {
			break
		}
		value := new(JSON)
		decodeJSON(iter, value)
		(*obj)[key] = value
	}
}

// anyencoder implements jsoniter.ValEncoder{} for Any{}.
type anyencoder struct{}

func (_ anyencoder) IsEmpty(ptr unsafe.Pointer) bool {
	a := (*Any)(ptr)
	return a.IsEmpty()
}

func (_ anyencoder) Encode(ptr unsafe.Pointer, stream *jsoniter.Stream) {
	encodeAny(stream, (*Any)(ptr))
}

// anydecoder implements jsoniter.ValDecoder{} for Any{}.
type anydecoder struct{}

func (_ anydecoder) Decode(ptr unsafe.Pointer, iter *jsoniter.Iterator) {
	decodeAny(iter, (*Any)(ptr))
}

// encodeAny encodes data in Any{} into Stream{}.
func encodeAny(stream *jsoniter.Stream, a *Any) {
	switch t := a.Type().(type) {
	case bool:
		stream.WriteBool(t)
	case int64:
		stream.WriteInt64(t)
	case uint64:
		stream.WriteUint64(t)
	case float64:
		stream.WriteFloat64(t)
	case string:
		stream.WriteString(t)
	case []*Any:
		encodeAnyArray(stream, t)
	case map[string]*Any:
		encodeAnyObject(stream, t)
	case nil:
		stream.WriteNil()
	default:
		panic(fmt.Sprintf("unexpected type %T", t))
	}
}

// decodeAny decodes next available value in Iterator{} into Any{}.
func decodeAny(iter *jsoniter.Iterator, a *Any) {
	switch t := iter.WhatIsNext(); t {
	case jsoniter.BoolValue:
		a.i = iter.ReadBool()
	case jsoniter.NumberValue:
		n := iter.ReadNumber()
		switch n := string(n); {
		case strings.HasPrefix(n, "-"):
			a.i, _ = strconv.ParseInt(n, 10, 64)
		case strings.Index(n, ".") >= 0:
			a.i, _ = strconv.ParseFloat(n, 64)
		default:
			a.i, _ = strconv.ParseUint(n, 10, 64)
		}
	case jsoniter.StringValue:
		a.i = iter.ReadString()
	case jsoniter.ArrayValue:
		var arr []*Any
		decodeAnyArray(iter, &arr)
		a.i = arr
	case jsoniter.ObjectValue:
		var obj map[string]*Any
		decodeAnyObject(iter, &obj)
		a.i = obj
	case jsoniter.NilValue:
		_ = iter.ReadNil()
		a.i = nil
	default:
		iter.ReportError("decodeAny", "BUG: unhandled value type")
		iter.Skip() // skip *whatever* this is.
	}
}

// anyarrayencoder implements jsoniter.ValEncoder{} for []*Any{}.
type anyarrayencoder struct{}

func (_ anyarrayencoder) IsEmpty(ptr unsafe.Pointer) bool {
	return *(*[]*Any)(ptr) == nil
}

func (_ anyarrayencoder) Encode(ptr unsafe.Pointer, stream *jsoniter.Stream) {
	encodeAnyArray(stream, *(*[]*Any)(ptr))
}

// anyarraydecoder implements jsoniter.ValDecoder{} for []*Any{}.
type anyarraydecoder struct{}

func (_ anyarraydecoder) Decode(ptr unsafe.Pointer, iter *jsoniter.Iterator) {
	decodeAnyArray(iter, (*[]*Any)(ptr))
}

// encodeAnyArray encodes data in []*Any{} into Stream{}.
func encodeAnyArray(stream *jsoniter.Stream, arr []*Any) {
	if arr == nil {
		// Write nil value.
		stream.WriteNil()
		return
	}

	stream.WriteArrayStart()
	for i := 0; i < len(arr)-1; i++ {
		encodeAny(stream, arr[i])
		stream.WriteMore()
	}
	encodeAny(stream, arr[len(arr)-1])
	stream.WriteArrayEnd()
}

// decodeAnyArray decodes data in Iterator{} into []*Any{}.
func decodeAnyArray(iter *jsoniter.Iterator, arr *[]*Any) {
	if arr == nil {
		// the ptr itself should never be nil.
		panic("BUG: decoder nil unsafe.Pointer")
	}

	switch iter.WhatIsNext() {
	// Allocate new arry slice.
	case jsoniter.ArrayValue:
		*arr = make([]*Any, 0, 32)

	// Don't set on nil value.
	case jsoniter.NilValue:
		_ = iter.ReadNil()
		return

	// Unusable type.
	default:
		iter.ReportError("decodeAnyArray", "invalid type")
		iter.Skip() // skip this array.
		return
	}

	for iter.ReadArray() {
		elem := new(Any)
		decodeAny(iter, elem)
		(*arr) = append((*arr), elem)
	}
}

// anyobjectencoder implements jsoniter.ValEncoder{} for map[string]*Any{}.
type anyobjectencoder struct{}

func (_ anyobjectencoder) IsEmpty(ptr unsafe.Pointer) bool {
	return *(*map[string]*Any)(ptr) == nil
}

func (_ anyobjectencoder) Encode(ptr unsafe.Pointer, stream *jsoniter.Stream) {
	encodeAnyObject(stream, *(*map[string]*Any)(ptr))
}

// anyobjectdecoder implements jsoniter.ValDecoder{} for map[string]*Any{}.
type anyobjectdecoder struct{}

func (_ anyobjectdecoder) Decode(ptr unsafe.Pointer, iter *jsoniter.Iterator) {
	decodeAnyObject(iter, (*map[string]*Any)(ptr))
}

// encodeAnyObject encodes data in map[string]*Any{} into Stream{}.
func encodeAnyObject(stream *jsoniter.Stream, obj map[string]*Any) {
	if obj == nil {
		// Write nil value.
		stream.WriteNil()
		return
	}

	var more bool

	stream.WriteObjectStart()
	for key, val := range obj {
		if more {
			stream.WriteMore()
		}
		stream.WriteObjectField(key)
		encodeAny(stream, val)
		more = true
	}
	stream.WriteObjectEnd()
}

// decodeAnyObject decodes data in Iterator{} into map[string]*Any{}.
func decodeAnyObject(iter *jsoniter.Iterator, obj *map[string]*Any) {
	if obj == nil {
		// the ptr itself should never be nil.
		panic("BUG: decoder nil unsafe.Pointer")
	}

	switch iter.WhatIsNext() {
	// Allocate new object map.
	case jsoniter.ObjectValue:
		*obj = make(map[string]*Any)

	// Don't set on nil value.
	case jsoniter.NilValue:
		_ = iter.ReadNil()
		return

	// Unusable type.
	default:
		iter.ReportError("decodeAnyObject", "invalid type")
		iter.Skip() // skip this object.
		return
	}

	for {
		key := iter.ReadObject()
		if key == "" {
			break
		}
		value := new(Any)
		decodeAny(iter, value)
		(*obj)[key] = value
	}
}
