package loosy_test

import (
	"encoding/json"
	"fmt"
	"testing"

	"codeberg.org/gruf/go-loosy"
)

var testData = toJSON(map[string]any{
	"key":    "value",
	"array":  []any{0, 1, 2, 3.3, 4.4, "hello", "world"},
	"name":   "test",
	"float":  3.14,
	"int":    -1,
	"uint":   1,
	"string": "hello world!",
	"object": map[string]any{"key": "value", "hello": "world", "int": 1},
	"null":   nil,
})

func TestObject(t *testing.T) {
	j := loosy.JSON{B: testData}

	obj, err := j.Object()
	if err != nil {
		t.Fatalf("failed marshaling test data: %v", err)
	}

	if str, _ := obj["key"].String(); str != "value" {
		t.Fatalf("failed to confirm object map value for key=key have=%v want=value", str)
	}

	if json, _ := obj["array"]; json == nil {
		t.Fatalf("failed to confirm object map value non-nil for key=array")
	} else {
		var ifaceArr []any

		if err := json.Value(&ifaceArr); err != nil {
			t.Fatalf("failed to unmarshal object map value as array for key=array")
		}

		expect := []any{0, 1, 2, 3.3, 4.4, "hello", "world"}

		if fmt.Sprintf("%v", ifaceArr) != fmt.Sprintf("%v", expect) {
			t.Fatalf("failed to confirm object map value for key=array have=%v want=%v", ifaceArr, expect)
		}
	}

	if json, _ := obj["object"]; json == nil {
		t.Fatalf("failed to confirm object map value non-nil for key=object")
	} else {
		var ifaceMap map[string]any

		if err := json.Value(&ifaceMap); err != nil {
			t.Fatalf("failed to unmarshal object map value as array for key=object")
		}

		expect := map[string]any{"key": "value", "hello": "world", "int": 1}

		if fmt.Sprintf("%v", ifaceMap) != fmt.Sprintf("%v", expect) {
			t.Fatalf("failed to confirm object map value for key=object have=%v want=%v", ifaceMap, expect)
		}
	}

	if str, _ := obj["name"].String(); str != "test" {
		t.Fatalf("failed to confirm object map value for key=name have=%v want=test", str)
	}

	if f, _ := obj["float"].Float(); f != 3.14 {
		t.Fatalf("failed to confirm object map value for key=float have=%v want=3.14", f)
	}

	if i, _ := obj["int"].Int(); i != -1 {
		t.Fatalf("failed to confirm object map value for key=int have=%v want=-1", i)
	}

	if u, _ := obj["uint"].Uint(); u != 1 {
		t.Fatalf("failed to confirm object map value for key=uint have=%v want=1", u)
	}

	if str, _ := obj["string"].String(); str != "hello world!" {
		t.Fatalf("failed to confirm object map value for key=string have=%v want=hello world!", str)
	}

	if !obj["null"].IsNull() {
		t.Fatal("failed to confirm object map nil value")
	}
}

func BenchmarkObjectUnmarshal(b *testing.B) {
	b.ResetTimer()
	b.RunParallel(func(pb *testing.PB) {
		for pb.Next() {
			var obj map[string]*loosy.JSON
			_ = loosy.Unmarshal(testData, &obj)
			_ = obj
		}
	})
}

func BenchmarkJsonMapUnmarshal(b *testing.B) {
	b.ResetTimer()
	b.RunParallel(func(pb *testing.PB) {
		for pb.Next() {
			var raw map[string]any
			_ = json.Unmarshal(testData, &raw)
			_ = raw
		}
	})
}

func toJSON(a any) []byte {
	b, err := loosy.Marshal(a)
	if err != nil {
		panic(err)
	}
	return b
}
